% a list of corners
% from image
clear all;
clc;
disp('Extrinsic parameters calibration procedure (SCABOR2)');

nrPoints=0;
xc=[];
yc=[];
ZWReal=[];
global rect1;

dir;

strImage = '';

%read the image pair
while isempty (strImage)
	strImage = input ('Image containing the targets full name plus extension: ','s');
end;

imTargets=imread(strImage);
hTargets=figure;
iptsetpref('ImshowBorder','loose');
imshow (imTargets);

Size_list=[];
Trash_list=[];
warning off;
if exist(['targets_size.mat'])
   string_save = ['load targets_size'];
   eval(string_save);
   fprintf(1,'\nLoading targets_size.mat ... done\n\n');
   is_ready='y';
   save_again=0;
else
    %Selecting the size of the targets
    is_ready='n';
    save_again=1;
    title('Click with the mouse in the middle of the smallest and bigest targets');
end; 

while is_ready=='n'
% manualy selection of the targets size from the image
    % mark a target
    [xman yman]=ginput3(1);
    xman=floor(xman);
    yman=floor(yman);
    posx=round(xman+0.5);
    posy=round(yman+0.5);
    h_line=line([posx-5, posx+5], [posy, posy]);
    set(h_line, 'color', 'g');
    h_line=line([posx, posx], [posy-5, posy+5]);
    set(h_line, 'color', 'g');

    % zoom in the target           
    zoomArea=25;
    zoomFactor=8;
    hZoom=figure;
    iptsetpref('ImshowBorder','tight');
    imZoom = imTargets(yman-zoomArea:yman+zoomArea, xman-zoomArea:xman+zoomArea);
    imZoomIn=imresize( imZoom , zoomFactor, 'bicubic');
    imshow(imZoomIn);
    title_name1='Drag with the mouse a rectangle around the target';
    set (hZoom,'Name',title_name1);
    % prepare the mouse events for rectangle region selection
    imHeight=size(imZoomIn,1);            
    figure(hZoom);
    button_pressed=0;
    ready_sel=0;
    set(gcf,'pointer','crosshair');
    set(gcf,'WindowButtonDownFcn','my_mouse_btn_down');
    set(gcf,'WindowButtonUpFcn','my_mouse_btn_up');
    set(gcf,'WindowButtonMotionFcn','my_mouse_btn_move');
    set(gcf,'Interruptible','on');

    while ready_sel==0
        pause(0.01);
        % wait until right mouse button is released
    end;
    
    if ready_sel==1
        %right mouse button is released
        twidth=rect1(3)/zoomFactor;
        theight=rect1(4)/zoomFactor;
        
        imHist=imZoomIn(rect1(2):rect1(2)+rect1(4),rect1(1):rect1(1)+rect1(3));
        bwt=hist_trashold(imHist,1); 
        Trash_list=[Trash_list bwt];
        
        tsize=(twidth+theight)/2;
        tsize=round(tsize);
        fprintf('\nSelected target size = %d\n',tsize);
        Size_list=[Size_list tsize];
        
        ready_sel=0;
        pause(2);
        delete(hZoom);
        figure(hTargets);
        iptsetpref('ImshowBorder','loose');
    end;

    is_ready='k';
    while is_ready~='y' & is_ready~='n' 
        is_ready=input('Finished selection? (Yes - y, No - n) ', 's');
    end;

end; %while

if save_again,
    string_save = ['save targets_size Size_list Trash_list'];
    eval(string_save);
    fprintf(1,'\nSaving targets_size.mat ... done\n\n');
end;

figure(hTargets);
iptsetpref('ImshowBorder','loose');
imshow (imTargets);

% reading the proper patterns
%Size_list;
min_size=min(Size_list);
max_size=max(Size_list);
med_size=round((max_size + min_size)/2);

disp('Patterns used for matching:');
min_file_str=[int2str(min_size) 'x' int2str(min_size) '.bmp']
med_file_str=[int2str(med_size) 'x' int2str(med_size) '.bmp']
max_file_str=[int2str(max_size) 'x' int2str(max_size) '.bmp']

imPattBig=double(im2bw(imread(max_file_str), 0.5));
imPattMedium=double(im2bw(imread(med_file_str), 0.5));
imPattSmall=double(im2bw(imread(min_file_str), 0.5));

Trash_list;
bwt=mean(Trash_list);
bwt=bwt/256;

fprintf('\n');
title_BWT=sprintf('Enter treshold value for image binarization [0..1]. Deafult value = %0.3f ',bwt);
title(title_BWT);

disp('B/W image tresholding value [0..1]):');
input_BWT=sprintf('BWT ([] = %0.3f) = ', bwt);

is_ready='n';
while is_ready=='n',
    BWT = input(input_BWT);
    if isempty(BWT), BWT = bwt; end; 
    I2=im2bw(imTargets, BWT);
    imTargetsbin=double(I2);
    h_I2=figure; imshow(I2);
    title_BWT=sprintf('Satisfied with the b/w trashold = %0.3f ?',BWT);
    title(title_BWT);

    is_ready='k';
    while is_ready~='y' & is_ready~='n' 
        is_ready=input('Satisfied with the b/w trashold? (Yes - y, No - n) ', 's');
    end
    bwt=BWT;
    delete(h_I2);
end

figure(hTargets);
title('Enter treshold value for the targets matching [0..1]. Deafult value = 0.25 ');
fprintf('\n');
disp('Treshold value for targets matching function [0..1]):');
THRESHOLD = input('THRESHOLD ([] = 0.25) = ');
if isempty(THRESHOLD), THRESHOLD = 0.25; end;

title('Select the region of intrest containing the targets by clicking its four extreme limits');
disp('Select the region of intrest containing the targets by clicking its four extreme limits');
% pause(2);
figure(hTargets);

min=10000000;
nrX=0;
hold on;

[xlimit ylimit]=ginput4(4);

y_max=size(imTargets, 1);
x_max=size(imTargets, 2);

[x_min pos]=min(xlimit);
[x_max pos]=max(xlimit);

[y_min pos]=min(ylimit);
[y_max pos]=max(ylimit);
hold off;

hold on;
rectangle('Position', [x_min y_min x_max-x_min y_max-y_min], 'Edgecolor','b' );

fprintf('\nMatching process in progress ');
tic; 
for yim=round(y_min):round(y_max),
     fprintf('.');
    for xim=round(x_min):round(x_max),
        
        match=sum(sum(abs(imTargetsbin(yim:yim+size(imPattBig, 1)-1, xim:xim+size(imPattBig, 1)-1)-imPattBig)))/(size(imPattBig, 1)^2);
        if match<THRESHOLD
            nrX=nrX+1;
            xsol(nrX)=xim;
            ysol(nrX)=yim;
            matchval(nrX)=match;
            patsize(nrX)=size(imPattBig, 1);
        end;
        
        match=sum(sum(abs(imTargetsbin(yim:yim+size(imPattMedium, 1)-1, xim:xim+size(imPattMedium, 1)-1)-imPattMedium)))/(size(imPattMedium, 1)^2);
        if match<THRESHOLD
            nrX=nrX+1;
            xsol(nrX)=xim;
            ysol(nrX)=yim;
            matchval(nrX)=match;
            patsize(nrX)=size(imPattMedium, 1);
        end;

        match=sum(sum(abs(imTargetsbin(yim:yim+size(imPattSmall, 1)-1, xim:xim++size(imPattSmall, 1)-1)-imPattSmall)))/(size(imPattSmall, 1)^2);
        if match<THRESHOLD
            nrX=nrX+1;
            xsol(nrX)=xim;
            ysol(nrX)=yim;
            matchval(nrX)=match;
            patsize(nrX)=size(imPattSmall, 1);
        end;
    end;
    
end;

%refine the 'x' array result
nrXUnique=0;
for i=1:nrX,
    
    if patsize(i)>0
        xmin=xsol(i)+patsize(i)/2-10;
        xmax=xsol(i)+patsize(i)/2+10;
        ymin=ysol(i)+patsize(i)/2-10;
        ymax=ysol(i)+patsize(i)/2+10;
        
        for j=i+1:nrX,
            if xsol(j)+patsize(j)/2>xmin & xsol(j)+patsize(j)/2<xmax & ysol(j)+patsize(j)/2>ymin & ysol(j)+patsize(j)/2<ymax 
                if (matchval(i)>matchval(j) & patsize(i)==patsize(j)) | patsize(i)<patsize(j)
                    xsol(i)=xsol(j);
                    ysol(i)=ysol(j);
                    matchval(i)=matchval(j);
                    patsize(i)=patsize(j);
                    patsize(j)=-1;
                else
                    patsize(j)=-1;
                end;
            end;
        end;
        
        nrXUnique=nrXUnique+1;
        xsolu(nrXUnique)=xsol(i);
        ysolu(nrXUnique)=ysol(i);
        patsizeu(nrXUnique)=patsize(i);
        
    end;

    
end;
toc;disp('sec.');
fprintf('\n\nThe number of distinct targets identified: %d\n', nrXUnique);
sizex=300;
sizey=300;

null_border=65;
for i=1:nrXUnique,
    
    match_map=zeros(sizey, sizex);
    
    analized_im=zeros(sizey, sizex);
    
    copy_analized_im=zeros(sizey, sizex);
    
    analized_im(1:sizey, 1:sizex)=double(imresize(imTargets(ysolu(i):ysolu(i)+patsizeu(i), xsolu(i):xsolu(i)+patsizeu(i)) , [sizey sizex], 'bicubic'));

    %determining the x position
    match_map=sum(analized_im);
    
    match_map(1:null_border)=0;
    match_map(sizex-null_border:sizex)=0;
    
    match_mapx=match_map;
        
    [max posx]=max(match_map);
    up=0;
    while 0.1*match_map(posx)>abs(match_map(posx)-match_map(posx+up))
        up=up+1;
    end;
    
    down=0;
    while 0.1*match_map(posx)>abs(match_map(posx)-match_map(posx-down))
        down=down+1;
    end;

    posx=posx+(up-down)/2;
    
    
    %determining the y position
    match_map=sum(analized_im');
    
    match_map(1:null_border)=0;
    match_map(sizey-null_border:sizey)=0;
    
    [max posy]=max(match_map);
    up=0;
    while 0.1*match_map(posy)>abs(match_map(posy)-match_map(posy+up))
        up=up+1;
    end;
    
    down=0;
    while 0.1*match_map(posy)>abs(match_map(posy)-match_map(posy-down))
        down=down+1;
    end;

    posy=posy+(up-down)/2;

%     figure
%     imshow(uint8(analized_im))
%     hold on;
% 
%     
%     h_line=line([posx-5, posx+5], [posy, posy]);
%     set(h_line, 'color', 'g');
%     h_line=line([posx, posx], [posy-5, posy+5]);
%     set(h_line, 'color', 'g');
    
    xsolu(i)=xsolu(i)-0.5+(posx-1)*patsizeu(i)/sizex;
    ysolu(i)=ysolu(i)-0.5+(posy-1)*patsizeu(i)/sizey;
end;

figure(hTargets);
for i=1:nrXUnique,
    posx=round(xsolu(i)+0.5);
    posy=round(ysolu(i)+0.5);
    h_line=line([posx-5, posx+5], [posy, posy]);
    set(h_line, 'color', 'g');
    h_line=line([posx, posx], [posy-5, posy+5]);
    set(h_line, 'color', 'g');
   
end;

    XWR=[];
    YWR=[];
    ZWR=[];

if exist(['targets_3D_coord.mat'])
   string_save = ['load targets_3D_coord'];
   eval(string_save);
   fprintf(1,'\nLoading targets_3D_coord.mat ... done\n\n');
   exist_3D=1;
else
    %Selecting the size of the targets
    exist_3D=0;
    fprintf('\nPlease enter the 3D coordinates for each target (current target is shown by a red cross)...\n');    
end; 

for i=1:nrXUnique,
    fprintf('\nTARGET %d\n', i);
%     % green the previous target
%     if i>1
%         posx=round(xsolu(i-1)+0.5);
%         posy=round(ysolu(i-1)+0.5);
%         h_line=line([posx-5, posx+5], [posy, posy]);
%         set(h_line, 'color', 'g');
%         h_line=line([posx, posx], [posy-5, posy+5]);
%         set(h_line, 'color', 'g');
%     end;
    
    % red the current target
    posx=round(xsolu(i)+0.5);
    posy=round(ysolu(i)+0.5);
    h_line=line([posx-5, posx+5], [posy, posy]);
    set(h_line, 'color', 'r');
    h_line=line([posx, posx], [posy-5, posy+5]);
    set(h_line, 'color', 'r');
%    h_text=text(posx+7, posy,int2str (i));
%    set(h_text,'color','r','Fontsize',8);
    figure(hTargets);
    title(['Current target (' num2str(i) ') marked with a red cross. Target correctly identified? (Yes - y, No - n) :']);
    % first the confirmation that the target is correct is required
    is_correct='k';
    while is_correct~='y' & is_correct~='n' 
        is_correct=input('Target correctly identified? (Yes - y, No - n) : ', 's');
    end;

    if is_correct=='n'

        is_manually='k';
   
        while is_manually~='y' & is_manually~='n' 
            is_manually=input('Correct manualy/ Reject ? (Correct - y, Reject - n) ', 's');
        end;
   
        %the user can choose it by mouse
        if is_manually=='y'
            figure(hTargets);
            
            posx=round(xsolu(i)+0.5);
            posy=round(ysolu(i)+0.5);
            h_line=line([posx-5, posx+5], [posy, posy]);
            set(h_line, 'color', 'y');
            h_line=line([posx, posx], [posy-5, posy+5]);
            set(h_line, 'color', 'y');

            [xman yman]=ginput3(1);
            xman=floor(xman);
            yman=floor(yman);
            zoomArea=15;
            zoomFactor=8;
            hZoom=figure;
            imshow(imresize(imTargets(yman-zoomArea:yman+zoomArea, xman-zoomArea:xman+zoomArea) , zoomFactor, 'bicubic'));
             title('Click with the mouse in the middle of the target');
            [xmans ymans]=ginput5(1);
            xsolu(i)=xman-zoomArea-0.5+(xmans-1)/zoomFactor;
            ysolu(i)=yman-zoomArea-0.5+(ymans-1)/zoomFactor;
            delete(hZoom);
            figure(hTargets);
        else
            %mark it as invalid
            posx=round(xsolu(i)+0.5);
            posy=round(ysolu(i)+0.5);
            h_line=line([posx-5, posx+5], [posy, posy]);
            set(h_line, 'color', 'y');
            h_line=line([posx, posx], [posy-5, posy+5]);
            set(h_line, 'color', 'y');
          	h_text=text(posx+9, posy,int2str (i));
            set(h_text,'color','y','Fontsize',8);

            xsolu(i)=0;
            ysolu(i)=0;
            continue;
        end;
        
    end;

    posx=round(xsolu(i)+0.5);
    posy=round(ysolu(i)+0.5);
    h_line=line([posx-5, posx+5], [posy, posy]);
    set(h_line, 'color', 'g');
    h_line=line([posx, posx], [posy-5, posy+5]);
    set(h_line, 'color', 'g');
    h_text=text(posx+9, posy,int2str (i));
    set(h_text,'color','g','Fontsize',10,'FontWeight','bold');
    title(['Enter 3D cooordinates of the current target (' num2str(i) ') :']);
    
%     if i>1
%         posx=round(xsolu(i-1)+0.5);
%         posy=round(ysolu(i-1)+0.5);
%         h_line=line([posx-5, posx+5], [posy, posy]);
%         set(h_line, 'color', 'g');
%         h_line=line([posx, posx], [posy-5, posy+5]);
%         set(h_line, 'color', 'g');
%     end;

    is_correct='n';
    
    if exist_3D,
        XWR(i)=XX(i,1);
        YWR(i)=XX(i,2);
        ZWR(i)=XX(i,3);
        fprintf('XWR(%d)=%d\n',i,XWR(i));
        fprintf('YWR(%d)=%d\n',i,YWR(i));
        fprintf('ZWR(%d)=%d\n',i,ZWR(i));
    else
    while is_correct=='n'
        XWR(i) = input('      X coordinate(mm) ');
        YWR(i) = input('      Y coordinate(mm) ');
        ZWR(i) = input('      Z coordinate(mm) ');
        
        is_correct='k';
        
        while is_correct~='y' & is_correct~='n' 
            is_correct=input('Accept coordinates? (Yes - y, No - n) ', 's');
        end;
     end;
    end;
        
end;

figure(hTargets);
title('All targets identified? (Yes - y, No - n) ');
fprintf('\n');
is_ready='k';
while is_ready~='y' & is_ready~='n' 
    is_ready=input('All targets identified? (Yes - y, No - n) ', 's');
end;

% select missed targets by mouse
while is_ready=='n'
    i=i+1;
    fprintf('\nTARGET %d\n', i)
    disp('Select missed target manually !');
    figure(hTargets);
    title('Select missed target manually by mouse !')
    [xman yman]=ginput3(1);
    xman=floor(xman);
    yman=floor(yman);
    zoomArea=15;
    zoomFactor=8;
    hZoom=figure;
    imshow(imresize(imTargets(yman-zoomArea:yman+zoomArea, xman-zoomArea:xman+zoomArea) , zoomFactor, 'bicubic'));
    title('Click with the mouse in the middle of the target !');
    [xmans ymans]=ginput5(1);
    xsolu(i)=xman-zoomArea-0.5+(xmans-1)/zoomFactor;
    ysolu(i)=yman-zoomArea-0.5+(ymans-1)/zoomFactor;
    delete(hZoom);
    figure(hTargets);
% Mark target on image
    posx=round(xsolu(i)+0.5);
    posy=round(ysolu(i)+0.5);
    h_line=line([posx-5, posx+5], [posy, posy]);
    set(h_line, 'color', 'g');
    h_line=line([posx, posx], [posy-5, posy+5]);
    set(h_line, 'color', 'g');
    h_text=text(posx+9, posy,int2str (i));
    set(h_text,'color','g','Fontsize',10,'FontWeight','bold');
% Enter 3D coordinates
    title(['Enter 3D cooordinates of the current target (' num2str(i) ') :']);
    
    is_correct='n';
    while is_correct=='n'
        XWR(i) = input('      X coordinate(mm) ');
        YWR(i) = input('      Y coordinate(mm) ');
        ZWR(i) = input('      Z coordinate(mm) ');
        
    is_correct='k';
        while is_correct~='y' & is_correct~='n' 
            is_correct=input('Accept coordinates ? (Yes - y, No - n) ', 's');
        end;
    end;
   
   figure(hTargets);
   title('All targets identified ? (Yes - y, No - n) ');
   fprintf('\n'); 
   is_ready='k';
   while is_ready~='y' & is_ready~='n' 
       is_ready=input('All targets identified ? (Yes - y, No - n) ', 's');
   end;
end;
nrXUnique=i;
warning on;
hold off;

x=[];
y=[];
XWReal=[];
YWReal=[];
ZWReal=[];
j=0;
%eliminate rejected points (those having 0 coordinate values)
for i=1:nrXUnique,
    if xsolu(i)~=0 & ysolu(i)~=0
        x=[x xsolu(i)];
        y=[y ysolu(i)];
        XWReal=[XWReal XWR(i)];
        YWReal=[YWReal YWR(i)];
        ZWReal=[ZWReal ZWR(i)];
        j=j+1;
    end;
end;

%sort points by spatial coordinates (Z and X) and plot them on the figure
XX=[];XXX=[];
XX=[XWReal' YWReal' ZWReal' x' y'];
%sort points by spatial coordinates (Z and X)
XXX=SORTROWS(XX,[3 1]);
XWReal=XXX(:,1)'
YWReal=XXX(:,2)'
ZWReal=XXX(:,3)'
x=XXX(:,4)'
y=XXX(:,5)'

% open the file 
name = input('\nPlease enter the name of file to save the 2D image coordinates and the 3D real world coordinates of the detected targets(without extension): ', 's');
string_save = ['save ' name '.mat XXX x y XWReal YWReal ZWReal'];
eval(string_save);
fprintf('Saving file %s.mat ... done.');

hTargets2=figure;
imshow(imTargets, 'truesize');
truesize(hTargets2,[720 1000]);
hold on;
title('Detected targets and their 2D image cordinates along with their 3D real world coordinates');

nrPoints=size(x,2);
offy=5;
if size(imTargets, 1)>520
    offxL=-500;
    offxR=30;
    pfact=10;
else
    offxL=-250;
    offxR=15;
    pfact=5;
end;
ftsz=13;

for i=1:nrPoints
    posx=round(x(i)+0.5);
    posy=round(y(i)+0.5);
    if posx<size(imTargets,2)/2
        offxL=offxL+ round(i/2)*pfact;
        offx=offxL;
    else
        offx=offxR;
    end;
    if mod(i,2)==1
        offy=offy-1;
        ftsz = ftsz-1;
        
    end;
    
    h_line=line([posx-5, posx+5], [posy, posy]);
    set(h_line, 'color', 'c');
    h_line=line([posx, posx], [posy-5, posy+5]);
    set(h_line, 'color', 'c');
    h_text1=text(posx+offx, posy+offy,['T' int2str(i) '(' int2str(posx) ',' int2str(posy) ')[' num2str(XWReal(i)/1000,3) ',' num2str(YWReal(i)/1000,3) ',' num2str(ZWReal(i)/1000,4) ']m'] );
    set(h_text1,'color','c','Fontsize',ftsz,'FontWeight','bold');
end;

hold off;


name=[name '.m'];
fid = fopen(name,'w');
%clear rejected points
fprintf(fid,'\n x = [');
fprintf(fid,'%f ', x);
fprintf(fid,']');

fprintf(fid,'\n y = [');
fprintf(fid,'%f ', y);
fprintf(fid,']');

fprintf(fid,'\n ZWReal = [');
fprintf(fid,'%f ',ZWReal);
fprintf(fid,']');

fprintf(fid,'\n YWReal = [');
fprintf(fid,'%f ',YWReal);
fprintf(fid,']');

fprintf(fid,'\n XWReal = [');
fprintf(fid,'%f ',XWReal);
fprintf(fid,']');
fclose(fid);

pause(1);

dir;
%open the intrinsic parameters for the camera
name_int = input('\nPlease enter the full name of file containing the intrinsic parameters of the camera (*.txt): ', 's');

[Fx, Fy, Pp]=load_int_txt(name_int);

F=-(Fx+Fy)/2;

gnewton

%save the extrinsic parameters for the camera
name_ext = input('\nPlease enter the name of file to save the extrinsic parameters of the camera (without extension): ', 's');

rc=rodrigues(Rc);
rw=rodrigues(Rw);
rcd=r2d(rc);
rwd=r2d(rw);

save_ext(name_ext,Tw,Rw,rw,rwd,Tc,Rc,rc,rcd);
